/*****************************************************************************
*
* Atmel Corporation
*
* File              : AVR465.c
* Compiler          : IAR EWAAVR 3.10C
* Revision          : $Revision: 1.31 $
* Date              : $Date: Friday, July 02, 2004 11:12:00 UTC $
* Updated by        : $Author: kmeyer $
*
* Support mail      : avr@atmel.com
*
* Supported devices : This example is written for the ATmega88. Firmware
*                     as such should fit in any other AVR with 8kB Flash.
*
* AppNote           : AVR465: Single-Phase Power/Energy Meter with
*                     Tamper Detection.
*
* Description       : Main file.
*
****************************************************************************/

/// DETAILS: ////////////////////////////////////////////////////////////////
//
//  Author............................ Kim Meyer, Atmel Finland Design Center
//  Compiler.................................... IAR Embedded Workbench 3.10C
//  Chip type (Flash/SRAM/EEPROM).................... ATmega88 (8kB/1kB/512B)
//  Program type................................................. Application
//  Clock frequency............................................. 4.000000 MHz
//  Average usage of program stack (RSTACK)......................0x0016 bytes
//  Average usage of data stack (CSTACK).........................0x00E0 bytes
//
//  Project Options for "Release":
//    General-Target-Processor Configuration........................ ATmega88
//    General-Target-Memory Model...................................... Small
//    General-Library Configuration-Enable bit definitions................ ON
//    General-Library Configuration-printf formater..................... Full
//    General-System Configuration-Data stack-Size..................... 0x0F0
//    General-System Configuration-Return address stack-Size.............. 24
//    ICCAVR-Optimisations........................................ Size, High
//    ICCAVR-Optimisations-Enabled-Common subexpression elimination....... NO
//    ICCAVR-Optimisations-Enabled-Function inlining...................... NO
//    ICCAVR-Optimisations-Enabled-Code motion........................... YES
//    ICCAVR-Optimisations-Enabled-Cross call............................ YES
//    ICCAVR-Optimisations-Enabled-Clustering of variables................ NO
//
//  Memory Usage for "Release":
//    Program (Code)...................................... 7743 bytes (94.5%)
//    Variables (Data)..................................... 709 bytes (69.2%)
//
/////////////////////////////////////////// All your base are belong to us //


/////////////////////////////////////////////////////////////////////////////
//  I N C L U D E S
/////////////////////////////////////////////////////////////////////////////

#include <iom88.h>
#include <stdio.h>
#include <math.h>           // This is to be able to calculate square roots
#include <inavr.h>          // This is for EEPROM routines
#include "defines.c"        // Definitions
#include "AVR465.h"            


/////////////////////////////////////////////////////////////////////////////
//  G L O B A L   V A R I A B L E S
/////////////////////////////////////////////////////////////////////////////

struct  AccDataStructure    { signed long PL, PN, U2, IL2, IN2;
                            } Accumulator, Sum;
struct	ResultDataStructure { float PL, PN, U, IL, IN;
                              char ADC0_Gain, ADC1_Gain;
                            } Result;
struct	CalDataStructure    { unsigned int PCC[3];
                              float ILG[3], ING[3];
                              float UG;
                              unsigned int MC, DPC;
                            } CalCoeff={{0,24576,49152},
                                        {0.01,0.001,0.0002},
                                        {0.01,0.001,0.0002},
                                        0.2,10000,100};
struct        SampleStructure {
                                signed int Fresh, Previous;
                                signed long Filtered,PreviousFiltered,Calibrated;
                              } Sample[3];

signed long   Temp[5];
unsigned int  TC1ext,TC1extTOP,TC1extRem;
unsigned int  DisplayCounter,DPCActive,DPCIncrement;
unsigned int  GainHysteresis;
unsigned char Index,ADC0_Gain=0xFF,ADC1_Gain=0xFF,Flags;
int           SampleCounter,RxBuffer;
float         Tnum;


/////////////////////////////////////////////////////////////////////////////
//  I N T E R R U P T   S E R V I C E   R O U T I N E S
/////////////////////////////////////////////////////////////////////////////

// Watchdog timeout handler
#pragma vector=WDT_vect
__interrupt void WFT_ISR(void)
{
  // put timeout handler here
}

// USART Receiver interrupt service routine
#pragma vector=USART_RX_vect
__interrupt void USART_RX_ISR(void)
{
  char Status,Data;
  
  Status=UCSR0A;
  Data=UDR0;
  if ((Status & ((1<<FE0)|(1<<UPE0)|(1<<DOR0)))==0)
  {
    RxBuffer=Data;
    Flags=Flags|KEY_PRESSED;
  }
  else
    RxBuffer=0x00;
}

// USART Transmitter interrupt service routine
#pragma vector=USART_TX_vect
__interrupt void USART_TX_ISR(void)
{
  // use this ISR if buffered data output is needed
}

// Timer 1 overflow interrupt service routine
#pragma vector=TIMER1_OVF_vect
__interrupt void TC1_OVF_ISR(void)
{
  // this should never occur
}

// Timer 1 output compare A ISR: output is set or cleared on compare match
#pragma vector=TIMER1_COMPA_vect
__interrupt void TC1_COMPA_ISR(void)
{
  // DPCIncrement should be zero when OC1A is cleared and one when OC1A is set
  DisplayCounter=DisplayCounter+DPCIncrement;

  // Prepare to set or clear output, if counter extension active
  if (TC1extTOP>0)
  {
    TC1ext++;               // Increase extended counter
    if (TC1ext<TC1extTOP)   // Extended counter not yet reached TOP
    {
      TCCR1A=0x80;          // CLEAR output on next match
      DPCIncrement=0;       // Do not increment display counter on next match
    }
    else                    // Extended counter has reached TOP
    {
      if (TC1ext>TC1extTOP)
      { // Output has now been set: start all over
        TC1ext=0;           // Clear extended counter
        OCR1A=0xFFFF;       // Count full 16-bit cycles until extended TOP
        TCCR1A=0x80;        // CLEAR output on next match
        DPCIncrement=0;     // Do not increment display counter on next match
      }
      else
      { // All 16-bit full cycles done: now do remainder and then set output
        OCR1A=TC1extRem;    // Remainder to count on last extended cycle
        TCCR1A=0xC0;        // SET output on next match
        DPCIncrement=1;     // Do increment display counter on next match
      }
    }
  }
  else
    TC1ext=0;
    
  if (CalCoeff.DPC==DisplayCounter)
  {
    DisplayCounter=0;
    PORTD=(PIND&DIRD)|DPP;  // Set DPP high, other outputs remain static
    DPCActive=1;            // Start counting active time for DPP/DPN
  }
}

// Timer 1 output compare B interrupt service routine
#pragma vector=TIMER1_COMPB_vect
__interrupt void TC1_COMPB_ISR(void)
{
  unsigned char Temp;
  
  // OC1A was set on OCR1A compare match, now clear output
  Temp=TCCR1A;        // Save current register (set/clear on next match?)
  TCCR1A=0x80;	      // Set mode: clear OC1A on compare match
  TCCR1C=0x88;	      // Now force a compare match to clear output
  TCCR1A=Temp;	      // Set mode: as was (set/clear on next match)
}

// ADC interrupt service routine with auto input scanning
#pragma vector=ADC_vect
__interrupt void ADC_ISR(void)
{
  signed int	TempI;
  signed long 	TempL;

  PORTB=(PINB&DIRB)|DUTY;               // For duty cycle monitoring
  
  // Sampled data from ADC:   ADCW      Range=[0000...03FF]
  // Copy of sampled data:    Sample    Range=[0000...03FF]
  // Scaled copy of data:     Temp      Range=[00000000...0003FB01]
  // DC filtered data:        Filtered  Range=[FFFE0280...0001FD80]
  
  // Save copy of fresh sample. Current samples are inverted; invert back.
  // Index variable keeps track of sampled channel, as follows:
  //
  // ISR:          |XXX|  |XXX|  |XXX|  |XXX|  |XXX|  |XXX|  |XXX|  |XXX|  |
  // Index/MUX:    XXX|000000|111111|222222|000000|111111|222222|000000|1111
  // Sampling:     |XXXXXX|000000|111111|222222|000000|111111|222222|000000|
  // Processing:   |XXXXXX|XXXXXX|000000|111111|222222|000000|111111|222222|
  // --------------+------+------+------+------+------+------+------+------+--
  // Conversion:   N     N+1    N+2    N+3    N+4    N+5    N+6    N+7    N+8
  //
  Sample[Index].Previous=Sample[Index].Fresh;	// x[n+1] <- x[n]
  if (0==Index)                                 // x[n] <- DATA
    Sample[Index].Fresh=ADC;            // save voltage sample as is
  else
    Sample[Index].Fresh=(0x03FF-ADC);   // save inverted current sample

  // Apply filter for DC offset removal. Must use long data types for
  // precision! The transfer function is as follows:
  //
  //	y[n] = 0.996*y[n-1] + 0.996*x[n] - 0.996*x[n-1]
  Sample[Index].PreviousFiltered=Sample[Index].Filtered;  // y[n] <- y[n-1]
  TempL=255*(long)Sample[Index].Filtered;
  TempL=TempL>>8;
  TempI=Sample[Index].Fresh-Sample[Index].Previous;
  TempL=TempL+255*(long)TempI;
  Sample[Index].Filtered=TempL;
  
  // Don't change gain settings until signals have settled
  if (GainHysteresis)
  {
    Flags=Flags&(0xFF-MOREGAIN0);
    Flags=Flags&(0xFF-MOREGAIN1);
    GainHysteresis--;
  }
  else
    // Gain control. Index=1 => ADC0, Index=2 => ADC1
    switch (Index)
    {
      case  1:  // Step down gain, if RAW sample has saturated at high limit
                if (Sample[Index].Fresh>SAT_HI)
                  Flags=Flags|LESSGAIN0;
                // Step down gain, if RAW sample has saturated at low limit
                if (Sample[Index].Fresh<SAT_LO)
                  Flags=Flags|LESSGAIN0;
                // Don't increase gain if FILTERED amplitude above threshold
                if (Sample[Index].Filtered>AMP_LO)
                  Flags=Flags&(0xFF-MOREGAIN0);
                break;
        case 2: // Step down gain, if RAW sample has saturated at high limit
                if (Sample[Index].Fresh>SAT_HI)
                  Flags=Flags|LESSGAIN1;
                // Step down gain, if RAW sample has saturated at low limit
                if (Sample[Index].Fresh<SAT_LO)
                  Flags=Flags|LESSGAIN1;
                // Don't increase gain if FILTERED amplitude above threshold
                if (Sample[Index].Filtered>AMP_LO)
                  Flags=Flags&(0xFF-MOREGAIN1);
    }
  
  // Apply phase calibration. Due to multiplexing, there will be a delay
  // between subsequent data chennels, as follows:
  //
  // 	Delay = [ 360 degrees * f(mains) ] / [ 3 * f(sampling) ]
  //
  // At 2400Hz sampling rate, this means that fresh samples from ADC0 
  // and ADC1, for example, will actually have a time difference of 
  // 1/2400 = 0.42ms, which corresponds to (50*360)/2400 = 7.5 degrees
  // in a 50Hz mains environment. In addition, current transformers in
  // the analogue front end may typically contribute up to six degrees
  // of additional phase lag. Phase distortions are calibrated using
  // linear interpolation, as follows:
  //
  //	Calibrated = x[n-1] + CalibCoefficient * ( x[n] - x[n-1] )
  //
  // Note that linear interpolation is the same as a fixed time delay.
  // This means that higher frequency componentes will not be adjusted
  // correctly. Typically, though, not enough energy is present in the
  // higher frequency components to cause this to degrade accuracy below
  // acceptable levels.
  //
  // Also note that the higher the sampling rate, the lower is the
  // calibration range in degrees. If the sampling rate is too high,
  // then the linear interpolation as such can not compensate for the
  // phase lag of some DC immuned current transformers. Such
  // transformers can have a phase lag in excess of six degrees.
  TempL=Sample[Index].Filtered-Sample[Index].PreviousFiltered;
  TempL=TempL*CalCoeff.PCC[Index];
  TempL=TempL>>16;
  Sample[Index].Calibrated=Sample[Index].Filtered-TempL;
  
  // Save measurement data for voltage, current and active power on all
  // channels. For active power measurements; accumulate instantaneous
  // power product. For current and voltage measurements: accumulate
  // square of samples
  //
  // Range check:
  //
  // 	Filtered data:		Filtered  Range=[FFFE0280...0001FD80]
  // 	Prescaled data (>>5):	TempX	  Range=[FFFFF014...00000FEC]
  // 	Multiplication result:	(n/a)	  Range=[FF027E70...00FD8190]
  // 
  // 	Prescaled data (>>6):	TempX	  Range=[FFFFF80A...000007F6]
  // 	Multiplication result:	(n/a)	  Range=[FFC09F9C...003F6064]
  // 
  // Assuming that sampled data is stuck at full-scale (which it can't
  // be, since the high-pass filter would bias such a signal to zero),
  // then for accumulation purposes the maximum number of samples that
  // can be integrated is:
  //
  // 	MaxSamples(presc=32) = 7FFFFFFFh / 00FD8190h = 0081h = 129d
  // 	MaxSamples(presc=64) = 7FFFFFFFh / 003F6064h = 0205h = 517d
  //
  // Note: Index=0 -> ADC2, Index=1 -> ADC0, Index=2 -> ADC1
  Temp[Index]=Sample[Index].Calibrated>>6;
  switch (Index)
  {
    case 0: // Index=0 => ADC2 (Voltage)
            Accumulator.U2=Accumulator.U2+(Temp[0]*Temp[0]);
            break;
    case 1: // Index=1 => ADC0 (Current L)
            Accumulator.IL2=Accumulator.IL2+(Temp[1]*Temp[1]);
            Temp[3]=Temp[0]*Temp[1];
            Accumulator.PL=Accumulator.PL+Temp[3];
            break;
    case 2: // Index=2 => ADC1 (Current N)
            Accumulator.IN2=Accumulator.IN2+(Temp[2]*Temp[2]);
            Temp[4]=Temp[0]*Temp[2];
            Accumulator.PN=Accumulator.PN+Temp[4];
            SampleCounter++;
            break;
  }
  if (SampleCounter==NMAX)
  {
    SampleCounter=0;
    if (Flags&GROGGY)
    {
      // Clear unsettled, accumulated data
      Accumulator.PL=0;
      Accumulator.PN=0;
      Accumulator.U2=0;
      Accumulator.IL2=0;
      Accumulator.IN2=0;
      Flags=Flags&(0xFF-GROGGY);          // Clear flag; we are OK now
    }
    else
      Flags=Flags|CYCLE_FULL;
  }
  
  // Maintain multiplexing of input channels
  if (++Index > 2)
    Index=0;
  ADMUX=ADC_VREF_TYPE+Index;		// Select next ADC input

  // Maintain display counter pulse, if activated. Toggle pulse as follows:
  //          ________
  // DPP:  __/        \___________________________________
  //                            ________
  // DPN:  ____________________/        \_________________
  //
  // TIME: --+--------+--------+--------+------------------>
  //         0      DPon    2*DP_ON  3*DP_ON      1T = 1/fs
  //
  // Use DP_ON constant to set pulse lentgh according to display counter specs
  if (DPCActive>0)
  {
    DPCActive++;
    switch (DPCActive)
    {
      case  DP_ON:    PORTD=(PIND&DIRD)&(0xFF-DPP);
                      break;
      case  2*DP_ON:  PORTD=(PIND&DIRD)|DPN;
                      break;
      case  3*DP_ON:  PORTD=(PIND&DIRD)&(0xFF-DPN);
                      DPCActive=0;
                      break;
    }
  }
  PORTB=(PINB&DIRB)&(0xFF-DUTY);
  asm("wdr");				// Watchdog reset
}                                       // End of ADC_ISR


/////////////////////////////////////////////////////////////////////////////
//  S U B F U N C T I O N S
/////////////////////////////////////////////////////////////////////////////

// Fundamental function for character-based input
int getchar(void)
{
  Flags=Flags&(0xFF-KEY_PRESSED);
  return RxBuffer;
}

// Fundamental function for character-based output (printf, etc)
int putchar(int Data)
{
  while ((UCSR0A & (1<<UDRE0))==0);
  UDR0=Data;
  return(Data);
}

// Calculate CRC16 Checksum based on initial checksum. Return new checksum.
unsigned int CRC(unsigned int chksum, unsigned char buffer)
{
  unsigned char j=8;
  
  chksum ^= (unsigned int)(buffer);
  do
  {
    if(chksum&0x8000)
      chksum=(chksum<<1) ^ CRC_POLYNOME;
    else 
      chksum=chksum<<1;
  }
  while(--j);
  return(chksum);
}

// Called after a device reset; initialises only registers which are required
// to have a non-default value
void Initialise(void)
{
  ACSR=0x80;      // Disable analogue comparator; save power

  DDRB=DIRB;
  DDRC=DIRC;
  DDRD=DIRD;

  TCCR1A=0x80;    // CLEAR output on next match (no EP output yet)
  TCCR1B=0x0D;	  // CTC Mode, Top=OCR1A, CLK/1024
  OCR1A=0xFFFF;   // Assume zero power; set max pulse interval
  OCR1B=0x00C3;   // Length of EP; T = (1024*195)/4000000 = 49.92ms
  TIMSK1=0x07;    // T/C1: Output compare match A & B + overflow

  UCSR0A=0x02;     // USART 2x speed enabled, Multi-processor mode disabled
  UCSR0B=0xD8;     // Rx and Tx enabled, Rx and tx interrupts enabled
  UCSR0C=0x06;     // Asynchronous, no parity, 8 data bits, 1 stop bit
  UBRR0=0x000C;    // f=4.000MHz, U2X=1: 38400 baud +/- 0.2%

  // ADC initialization:
  // Enable ADC interrupt and use internal voltage reference with stabilising
  // capacitor on AREF. Set multiplexer input to first channel and do not set
  // ADC to High Speed Mode. Set Auto Trigger Source to Free Running Mode.
  //
  //  fCLK | ADPS | Prescaler | ADCCLK | Sampl.Rate
  // ------+------+-----------+--------+-----------
  // 4.000 |  111 |       128 | 31250  |       2403

  ADMUX=ADC_VREF_TYPE;
  ADCSRA=0xEF;    // Enable ADC, start, auto trigger, int enable, presc = 128

  Flags=Flags|DISPLAY;            // Set flag
  Flags=Flags|GROGGY;             // Just woke up, hold on for a while...
  TC1extTOP=0xFFFF;               // Assume zero power; set max count
  DPCIncrement=0;                 // No increment; no displ counter pulses yet

  SampleCounter=-1200+NMAX;       // Allow HPF to settle to within 1%
  PORTC=(PINC&DIRC)|EARTH|REVDIR; // Set signals high to start with
}

void InitGainControl(void)
{
  // The LessGain flags will be set when any sample reaches high limit
  Flags=Flags&(0xFF-LESSGAIN0);
  Flags=Flags&(0xFF-LESSGAIN1);
  
  // The MoreGain flags will be cleared when any sample reaches threshold
  Flags=Flags|MOREGAIN0;
  Flags=Flags|MOREGAIN1;
}

void ReadCalibration(void)
{
  // Remember that interrupts must be disabled when accessing EEPROM!
  //
  // ADDR | +0x00 | +0x01 | +0x02 | +0x03 | +0x04 | +0x05 | +0x06 | +0x07
  // -----+-------+-------+-------+-------+-------+-------+-------+-------
  // 0x00 | PCC0H | PCC0L | PCC1H | PCC1L | PCC2H | PCC2L | ILG0H | ILG0L
  // 0x08 | ILG1H | ILG1L | ILG2H | ILG2L | ING0H | ING0L | ING1H | ING1L
  // 0x10 | ING2H | ING2L |   UGH |   UGL |   MCH |   MCL |  DPCH |  DPCL
  // 0x18 |       |       |       |       | CRCWH | CRCWL | CRC16 | CRC16
  //
  // FEATURE: DO NOT FILL EEPROM WITH ZEROS, SINCE IT WILL GIVE A VALID
  // CHECKSUM BUT INVALID CALIBRATION COEFFICIENTS!
  
  unsigned char n=0,Hi,Lo;
  unsigned int crc=0;
  
  // Calculate checksum for first 24 bytes of EEPROM  
  for (n=0;n<0x18;n++)
  {
    __EEGET(Hi,n);
    crc = CRC(crc,Hi);
  }
  
  // Write checksum to EEPROM if CRCW = 0x4357
  __EEGET(Hi,0x1C);
  if (0x43==Hi)
  {
    __EEGET(Lo,0x1D);
    if (0x57==Lo)
    {
      __EEPUT(0x1C,0x00);
      __EEPUT(0x1D,0x00);
      __EEPUT(0x1E,(crc>>8));
      __EEPUT(0x1F,(crc&0xFF));
    }
  }
  
  // Read calibration coefficients from EEPROM, if checksum holds
  __EEGET(Hi,0x1E);
  __EEGET(Lo,0x1F);
  if (crc-(Hi<<8)-Lo)
  {
    // put error handler here
  }
  else
  {
    for (n=0;n<3;n++)
    {
      __EEGET(Hi,0x00 + n*2);
      __EEGET(Lo,0x01 + n*2);
      CalCoeff.PCC[n]=(Hi<<8) + Lo;
  
      __EEGET(Hi,0x06 + n*2);
      __EEGET(Lo,0x07 + n*2);
      CalCoeff.ILG[n]=1.0 / ((Hi<<8) + Lo);
  
      __EEGET(Hi,0x0C + n*2);
      __EEGET(Lo,0x0D + n*2);
      CalCoeff.ING[n]=1.0 / ((Hi<<8) + Lo);
    }
  
    __EEGET(Hi,0x12);
    __EEGET(Lo,0x13);
    CalCoeff.UG=(unsigned int)((Hi<<8) + Lo) / 65536.0;
    
    __EEGET(Hi,0x14);
    __EEGET(Lo,0x15);
    CalCoeff.MC=(Hi<<8) + Lo;
    
    __EEGET(Hi,0x16);
    __EEGET(Lo,0x17);
    CalCoeff.DPC=(Hi<<8) + Lo;
  }

  // Save clock cycles; calculate numerator now, rest when adjusting EP T/C:
  //
  //   Tnum = (3600 x fCLK)/(Prescaler x P[kW])
  //        = (3600 x 4000000 x 1000)/(1024 x MC)
  //        = 14 062 500 000 / MC
  Tnum=14062500000.0/CalCoeff.MC;
}

void SetGain(unsigned char Channel, unsigned char Level)
{
  //  ADC |        |      |    PD    |   PORTD  |   DDRD  
  //  Ch. |  Range | Gain | 76543210 | 76543210 | 76543210
  // -----+--------+------+----------+----------+----------
  //   0  |    Low |    1 | xxxxLLxx | xxxx00xx | xxxx11xx
  //   0  | Medium |    8 | xxxxHLxx | xxxx10xx | xxxx11xx
  //   0  |   High |   64 | xxxxLHxx | xxxx01xx | xxxx11xx
  //   1  |    Low |    1 | xxLLxxxx | xx00xxxx | xx11xxxx
  //   1  | Medium |    8 | xxHLxxxx | xx10xxxx | xx11xxxx
  //   1  |   High |   64 | xxLHxxxx | xx01xxxx | xx11xxxx
  
  #define   GC0   0x0C
  #define   GC1   0x30
  
  switch (Level)
  {
    case LOW:	switch (Channel)
                {
                  case ADC0:  // set gain control pins LOW, only
                              PORTD=(PIND&DIRD)&(0xFF-GC0);
                              ADC0_Gain=LOW;
                              break;
                  case ADC1:  // set both gain control pins LOW
                              PORTD=(PIND&DIRD)&(0xFF-GC1);
                              ADC1_Gain=LOW;
                              break;
                }
                break;
    case MEDIUM:  switch (Channel)
                  {
                    case ADC0:  // set PD2 LOW and PD3 HIGH
                                PORTD=((PIND&DIRD)&(0xFF-GC0))|(1<<PD3);
                                ADC0_Gain=MEDIUM;
                                break;
                    case ADC1:  // set PD4 LOW and PD5 HIGH
                                PORTD=((PIND&DIRD)&(0xFF-GC1))|(1<<PD5);
                                ADC1_Gain=MEDIUM;
                                break;
                  }
                  break;
    case HIGH:	  switch (Channel)
                  {
                    case ADC0:  // set PD2 HIGH and PD3 LOW
                                PORTD=((PIND&DIRD)&(0xFF-GC0))|(1<<PD2);
                                ADC0_Gain=HIGH;
                                break;
                    case ADC1:  // set PD4 HIGH and PD5 LOW
                                PORTD=((PIND&DIRD)&(0xFF-GC1))|(1<<PD4);
                                ADC1_Gain=HIGH;
                                break;
                  }
                  break;
  }
  GainHysteresis=GAIN_HOLD;
}

void SetPulse(float Power)
{
  // Set output frequency of energy pulse as follows:
  //
  //       fCLK         3600    
  //  T = ------ x ------------ 
  //      Presc.    P[kW] x MC  
  //
  // Defaults are 4MHz system clock and 1024 prescaler. The default meter
  // constant, MC, is 10000 pulses/kWh.
  //
  // The energy pulse output is controlled by hardware T/C, which is driven 
  // by a prescaled system clock. Using the above settings, the T/C output is
  // good for 0.2% (i.e. 1/500) accuracy, since:
  //
  //  Tmin = 14062500000 / (265V*10A*10000imp/kWh) = 530
  //
  // Since T/C1 is only 16 bit in length, an extension to the counter is 
  // stored in TC1ext. The extended counter has a maximum pulse period of:
  //
  //         Presc
  // Tmax = ------- x ( FFFF x FFFF + FFFF ) = >1Ms = >305h = >12d
  //          fCLK
  //
  // ...but, in practice active power is either zero or above I_MIN x Unom.
  // Assuming I_MIN = 0.002 and Unom = 230V, and high meter constant:
  //
  //          4000000                  3600
  //  Tmax = --------- x ------------------------------- = 3057s = <1h
  //            1024      0.002A x 230V x 10000 imp/kWh

  #define Tmax  0xFFFE0000
  
  float T;
  unsigned int CtrTemp,ExtCtrTemp;
  char Mask;

  Power=fabs(Power);      // Pulse interval contains no sign information
  if (Power<0.0001)       // Divide by zero is not defined in this universe
    T=Tmax;
  else
    T=Tnum/Power;         // Precalculated numerator assumes power in watts
  
  if (T<(2*OCR1B))
    T=2*OCR1B;                      // Min interval = twice pulse high time

  if (T>Tmax)
  {
    ExtCtrTemp=0xFFFF;              // Max count
    CtrTemp=0xFFFF;                 // Max count
  }
  else
  {
    ExtCtrTemp=T/0xFFFF;            // Number of 16-bit timer cycles
    CtrTemp=T-(ExtCtrTemp*0xFFFF);  // Ticks in single/last cycle
  }

  Mask=TIMSK1;
  TIMSK1=0x00;                      // Disable T/C1 interrupts for a moment

  if (TC1extTOP>0)
  { // Currently timer is in the middle of an extended count
    TC1extTOP=ExtCtrTemp;         // Set new TOP for extended count
    if (TC1ext>TC1extTOP)
      TC1ext=TC1extTOP;           // Fit counter value in new frame
    if (TC1ext==TC1extTOP)
    { // Last extended count cycle
      OCR1A=CtrTemp;              // Set TOP value for last cycle
      TCCR1A=0xC0;                // Set mode: set OC1A on next compare match
      DPCIncrement=1;             // Do increment display counter on next match
    }
    else
    { // At least one more extended cycle to go
      OCR1A=0xFFFF;               // Set TOP value for last cycle
      TCCR1A=0x80;                // Set mode: clear OC1A on next match
      DPCIncrement=0;             // Do not increment displ cntr on next match
    }
  }
  else
  { // Currently timer runs unextended
    OCR1A=CtrTemp;                // Set new TOP value
    if (ExtCtrTemp>0)
    { // Next cycle must run extended
      TC1extTOP=ExtCtrTemp;       // Set new TOP for extended count
      TC1ext=TC1extTOP;           // Pretend we are at end of extended cycle    
    }
    TCCR1A=0xC0;                  // Set mode: set OC1A on next compare match
    DPCIncrement=1;               // Do increment display counter on next match
  }
  TC1extRem=CtrTemp;              // Set TOP value for future last cycles
  if (TCNT1>=OCR1A)
    TCNT1=(OCR1A-1);              // Rewind, if counter is larger than TOP

  TIMSK1=Mask;                    // Restore T/C1 interrupts
}


/////////////////////////////////////////////////////////////////////////////
//  M A I N
/////////////////////////////////////////////////////////////////////////////

void main(void)
{
  unsigned long	TempUL;
  
  // Check source of reset
  if (MCUSR & 1)               // Power-on Reset
  {
    // put POR handler here, if required
  }
  else if (MCUSR & 2)          // External Reset
  {
    // put external reset handler here, if required
  }
  else if (MCUSR & 4)          // Brown-Out Reset
  {
    // put BOR handler here, if required
  }
  else                          // Watchdog Reset
  {
    // put watchdog reset handler here, if required
  };
  MCUSR=0;

  Initialise();           // set up I/O registers, flags & variables
  ReadCalibration();      // interrupts must be disabled when accessing EEPROM
  InitGainControl();      // configure variable gain
  SetGain(ADC0,LOW);      // start from low gain
  SetGain(ADC1,LOW);      // start from low gain
  
  asm("wdr");                                 // reset WD counter
  WDTCSR |= (1<<WDCE) | (1<<WDE);             // enable watchdog interrupt 
  WDTCSR = (1<<WDE) | (1<<WDP2) | (1<<WDP1);  // set 128K cycles (~1.0 s)
  asm("sei");                                 // allow all enabled interrupts
  
  printf("\n\r");
  printf("AVR465\n\r");
  printf("\n\r");
  printf("PCC:   0x%04X   0x%04X   0x%04X\n\r",CalCoeff.PCC[0],CalCoeff.PCC[1],CalCoeff.PCC[2]);
  printf("ILG: %8.6f %8.6f %8.6f\n\r",CalCoeff.ILG[0],CalCoeff.ILG[1],CalCoeff.ILG[2]);
  printf("ING: %8.6f %8.6f %8.6f\n\r",CalCoeff.ING[0],CalCoeff.ING[1],CalCoeff.ING[2]);
  printf("UG:  %8.6f\n\r",CalCoeff.UG);
  printf("MC:  %8u\n\r",CalCoeff.MC);
  printf("DPC: %8u\n\r",CalCoeff.DPC);
  printf("\n\r");
  
  while (1)
  {
    if (Flags&KEY_PRESSED)
    {
      // Insert key service functions here
      switch (getchar())
      {
        case 'd':
        case 'D': if (Flags&DISPLAY)
                    Flags=Flags&(0xFF-DISPLAY);
                  else
                    Flags=Flags|DISPLAY;
                  break;
      }
    }

    // When a full cycle of data has been accumulated; make a copy of accumulated
    // data before it is overwritten by next sampling instance. Process accumulated
    // data as follows:
    //
    // Active power:	P = (accumulated data) / (number of samples)
    // Voltage:		U = SQRT [ (accumulated data) / (number of samples) ]
    // Current:		I = SQRT [ (accumulated data) / (number of samples) ]
    // Apparent power:	S = U * I
    //
    // All data processing must be complete before next accumulation cycle is
    // completed.
    if (Flags&CYCLE_FULL)
    {
      Flags=Flags&(0xFF-CYCLE_FULL);
      
      // Make a copy of accumulated data before it is overwritten
      Sum=Accumulator;
      
      // Clear accumulated data
      Accumulator.PL=0;
      Accumulator.PN=0;
      Accumulator.U2=0;
      Accumulator.IL2=0;
      Accumulator.IN2=0;
      
      // Save gain settings that was used with this data
      Result.ADC0_Gain=ADC0_Gain;
      Result.ADC1_Gain=ADC1_Gain;

      // Set gain for next set of measurements
      if (Flags&LESSGAIN0)
        switch (ADC0_Gain)
        {
          case MEDIUM:  SetGain(ADC0,LOW);
                        break;
          case HIGH:    SetGain(ADC0,MEDIUM);
                        break;
        }
      if (Flags&MOREGAIN0)
        switch (ADC0_Gain)
        {
          case LOW:     SetGain(ADC0,MEDIUM);
                        break;
          case MEDIUM:  SetGain(ADC0,HIGH);
                        break;
        }
      if (Flags&LESSGAIN1)
        switch (ADC1_Gain)
        {
          case MEDIUM:  SetGain(ADC1,LOW);
                        break;
          case HIGH:    SetGain(ADC1,MEDIUM);
                        break;
        }
      if (Flags&MOREGAIN1)
        switch (ADC1_Gain)
        {
          case LOW:     SetGain(ADC1,MEDIUM);
                        break;
          case MEDIUM:  SetGain(ADC1,HIGH);
                        break;
        }
      
      InitGainControl();

      // Calculate. Use accumulated data to derive active power and current
      // for live and neutral wires. Also derive voltage between live and
      // neutral wire.
      
      // Start by adding offset to accumulated data. This will improve non-
      // linearities at small amplitudes. Provided that the offset is
      // sufficiently small, that is!
      if (Sum.PL>0)
        Result.PL=Sum.PL*NORM+OFFSET;
      else
        Result.PL=Sum.PL*NORM-OFFSET;
      if (Sum.PN>0)
        Result.PN=Sum.PN*NORM+OFFSET;
      else
        Result.PN=Sum.PN*NORM-OFFSET;
      TempUL=(Sum.U2+OFFSET)*NORM;
      Result.U=sqrt(TempUL);
      TempUL=(Sum.IL2+OFFSET)*NORM;
      Result.IL=sqrt(TempUL);
      TempUL=(Sum.IN2+OFFSET)*NORM;
      Result.IN=sqrt(TempUL);

      // Calibrate. Take measurement results and adjust them according to
      // characteristics of this particular meter.
      Result.PL=Result.PL*CalCoeff.ILG[Result.ADC0_Gain];
      Result.IL=Result.IL*CalCoeff.ILG[Result.ADC0_Gain];
      Result.PN=Result.PN*CalCoeff.ING[Result.ADC1_Gain];
      Result.IN=Result.IN*CalCoeff.ING[Result.ADC1_Gain];
      Result.PL=Result.PL*CalCoeff.UG;
      Result.PN=Result.PN*CalCoeff.UG;
      Result.U=Result.U*CalCoeff.UG;

      // Anti-creep: reset power readings when current below threshold.
      // This is to make sure no pulses are emitted when signal is too low.
      // Don't want to bill the customer for noise in input channels...?
      if (Result.IL<I_MIN)
        Result.PL=0.0;
      if (Result.IN<I_MIN)
        Result.PN=0.0;

      // Tampering logic activated when either active power above threshold.
      // This is to prevent tamper indicators to go on and off at very low
      // power readings.
      if ((fabs(Result.PL)>P_MIN) | (fabs(Result.PN)>P_MIN))
      {
        // Check for earth fault, i.e. verify that all current which is going
        // into the load via the live wire also returns via the neutral wire.
        // Set tamper indicator as follows:
        //
        //    EARTH | Condition
        //   -------+-------------------------------------------
        //     Low  | P(L) and P(N) are almost same in magnitude
        //    High  | P(L) and P(N) are not same in magnitude
        //
        // It doesn't matter if the user swappes live and neutral wires.
        // Use the larger power reading for billing (energy pulse output).
        if (fabs(Result.PL)>fabs(Result.PN))
        { // |P(L)| > |P(N)|
          if (fabs(Result.PL)>(1.05*fabs(Result.PN)))
            PORTC=(PINC&DIRC)|EARTH;        // Magn. diff. > 5% => Set high
          else
            PORTC=(PINC&DIRC)&(0xFF-EARTH); // Magn. diff. < 5% => Set low
          SetPulse(Result.PL);        // Use Live wire measurements
        }
        else
        { // |P(L)| < |P(N)|
          if (fabs(Result.PN)>(1.05*fabs(Result.PL)))
            PORTC=(PINC&DIRC)|EARTH;        // Magn. diff. > 5% => Set high
          else
            PORTC=(PINC&DIRC)&(0xFF-EARTH); // Magn. diff. < 5% => Set low
          SetPulse(Result.PN);        // Use Neutral wire measurements
        }

        // Verify that current flows in the expected direction. Both live and
        // neutral current are positive if design is wired as expected. Set
        // tamper indicator as follows:
        //
        //   REVDIR | Condition
        //   -------+--------------------------------------
        //     Low  | P(L) is positive AND P(N) is positive
        //    High  | P(L) is negative OR P(N) is negative
        //
        // Measurement accuracy is not compromised by sign of current.
        if ((Result.PL<0) | (Result.PN<0))
          PORTC=(PINC&DIRC)|REVDIR;         // Current flow NOK => Set high
        else
          PORTC=(PINC&DIRC)&(0xFF-REVDIR);  // Current flow OK => Set low
      }
      else
      {
        // Tampering logic is disabled for now
        PORTC=(PINC&DIRC)&(0xFF-EARTH);
        PORTC=(PINC&DIRC)&(0xFF-REVDIR);
        SetPulse(Result.PL);                // Default
      }
      
      // Display measurement results. Connect TxD and RxD to any terminal
      // software or application. If properly calibrated, the output will
      // show active power in watts, current in amps and voltage in volts.
      if (Flags&DISPLAY)
      {
        printf(" %12.3f, %12.3f",Result.PL,Result.PN);
        printf(", %10.5f, %10.5f",Result.IL,Result.IN);
        printf(", %8.3f",Result.U);
        printf("\n\r");
      }
    }
  }
}
